-- name=m_quests
--[[

  Типы квестов (QT = quest type):
  - QT_StageCompleted        -- линейная генерация
  - QT_StageListCompleted
  - QT_ReachMaxLevelStages
  - QT_StageItemsCollected
  - QT_SpellCompleted        -- динамическая генерация
  - QT_BuildConstructed
  - QT_MineConstructed
  - QT_MarketConstructed
  - QT_MineResourceCollected -- динамическая генерация
  - QT_FragmentCollected

  Условия старта квеста (QSC = quest start condition):
  - QSC_Empty
  - QSC_QuestCompleted
  - QSC_StageCompleted
  - QSC_SpellEnabled
  - QSC_SpellCompleted
  - QSC_BuildConstructed
  - QSC_MineConstructed
  - QSC_MineResourceCollected
  - QSC_MineResourceFilled
  - QSC_FragmentCollected
]]
--******************************************************************************
--******************************************************************************
function public.Init ()
  
  private.caches = {
    current_quests  = nil,
    current_expired = true,
    current_stage   = nil
  };

  private.default = {};
  private.default.complete_stage_quest_name = "qst_complete_stage_";
  private.default.complete_spell_quest_name = "qst_complete_";
  private.default.mine_collect_quest_name   = "qst_mine_collect_resources_";

  --- генерация динамических квестов ---
  private.ready_mines_list    = private.GetReadyMines();
  private.enabled_spells_list = private.GetEnabledSpells();

  private.InitSubscribes();
  
end;
--******************************************************************************
--******************************************************************************
function private.InitSubscribes ()
  
  mediator.Subscribe( "Model_Stage_StageWon",        function ( params ) 

    private.EventParser( "StageCompleted",      params );
    private.EventParser( "StageListCompleted",  params );
    private.EventParser( "ReachMaxLevelStages", params );

  end );

  mediator.Subscribe( "Model_Stage_QuestItemDropped",   function ( params ) 
  
    private.EventParser( "StageItemsCollected", params ); 
  
  end );

  mediator.Subscribe( "Model_Builds_BuildConstructed",  function ( params ) 
  
    private.EventParser( "BuildConstructed", params ); 
  
  end );

  mediator.Subscribe( "Model_Mines_MineStateChanged",   function ( params ) 

    if     ( params.state == "active" ) then

      private.EventParser( "MineConstructed",       params ); 
      private.EventParser( "MineResourceCollected", params ); 

    elseif ( params.state == "ready" ) then

      private.EventParser( "MineResourceFilled",    params ); 

    end;

  end );

  mediator.Subscribe( "Model_Spells_SpellEnabled",      function ( params ) 
  
    private.EventParser( "SpellEnabled", params ); 
  
  end );

  mediator.Subscribe( "Model_Spells_SpellCompleted",    function ( params ) 
  
    private.EventParser( "SpellCompleted", params ); 
  
  end );

  mediator.Subscribe( "Model_Items_ItemObtain",       function ( params ) 

    if ( params.type == "fragment" ) then
      
      private.EventParser( "FragmentCollected", params );
      
    end;

  end );

  mediator.Subscribe( "Model_Market_Constructed",       function ( params ) 

    private.EventParser( "MarketConstructed", params );

  end );
  
end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function public.GetCompleteQuestCount ()

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "quests_completed"
  };
  
  local quest_completed = m_storage.Execute( query ).result;

  if ( not quest_completed ) then return 0;
  else                            return #quest_completed;
  end;

end;
--------------------------------------------------------------------------------
function public.GetCurrentQuestsInfo ( section_name )

  return private.GetCurrentQuestsInfo( nil, "external", section_name );

end;
--------------------------------------------------------------------------------
function public.IsStageItemActive ( item_name, stage_name, source, section )

  return private.IsStageItemActive( item_name, stage_name, source, section );

end;
--------------------------------------------------------------------------------
function public.IsQuestCurrent ( quest_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return ( m_storage.Execute( query ).result ~= nil );

end;
--------------------------------------------------------------------------------
function public.IsQuestComplete ( quest_name )

  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_completed",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return ( m_storage.Execute( query ).result ~= nil );

end;
--------------------------------------------------------------------------------
function public.GetCurrentQuestNameByItem ( item_type, item_name )

  local quests = public.GetCurrentQuestsInfo( m_map.Section_Current_Name() );

  for i, quest in pairs( quests ) do

    local rewards = quest.reward or {};

    for r, reward in pairs( rewards ) do

      local same_type = ( reward.type == item_type );
      local same_name = ( reward.name == item_name );

      if ( same_type and same_name ) then 

        return quest.name;

      end;

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.GetQuestByTarget ( target_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "target.name",
    criteria   = { like = target_name }
  };

  return m_storage.Execute( query ).result;

end;
--------------------------------------------------------------------------------
--function ... () end;
--------------------------------------------------------------------------------
function private.GetCurrentQuestsInfo ( quest_type, source, section_name )

  section_name = section_name or private.current_section;

  if ( section_name ~= private.current_section ) then

    private.caches.current_expired = true;

  end;

  private.current_section = section_name;

  if 
    ( not quest_type ) and 
    ( not private.caches.current_expired ) and 
    ( private.caches.current_quests ) and 
    (( source == "external" ) or ( source == "map" ) 
    --[[or ( source == "stage" )]]) then

    return private.caches.current_quests;

  end;

  local quests_out = {};
  
  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "quests_current"
  };
  
  local quests_current = m_storage.Execute( query ).result;
  
  private.GetSavedQuests(   quest_type, section_name, source, quests_current, quests_out );
  private.GetPregiveQuests( quest_type, section_name, source, quests_current, quests_out );
  private.GetLinearQuests(  quest_type, section_name, source, quests_out );
  private.GetDynamicQuests( quest_type, section_name, source, quests_out );
  
  if ( not quest_type ) then

    private.caches.current_quests = quests_out;
    private.caches.current_expired = false;

  end;
  
  return quests_out;
  
end;
--******************************************************************************
function private.GetSavedQuests ( quest_type, section_name, source, quests_current, quests_out )

  -- AL> Квесты из сэйва
  for q, quest_current in pairs( quests_current ) do

    local quest_data = public.GetQuestData( quest_current.name );

    if ( not quest_type ) or ( quest_type == quest_data.target.type ) then

      if ( not quest_current.count ) and ( not quest_current.list ) then      

        quest_current.list = {};

      end;

      if 
      source == "stage" or
      source == "event_parser" or
      source == "map" or      --( source == "map" and ( not quest_current.pregive )) or
      ( source == "external" and ( not quest_data.hidden ) and ( not quest_current.pregive ) 
        and ( section_name == quest_data.section ))
      then

        if ( quest_data.reward ) then

          if ( #quest_data.reward == 0 ) then

            local rewards = {};
            table.insert( rewards, quest_data.reward );
            quest_data.reward = rewards;

          end;

          for r = 1, #quest_data.reward do
            quest_data.reward[ r ].content = m_content.GetContent( 
              quest_data.reward[ r ].type, 
              quest_data.reward[ r ].name 
            );
          end;

        end;

        table.insert( quests_out,
        {
          name         = quest_current.name,
          pregive      = quest_current.pregive,
          count        = quest_current.count,
          list         = quest_current.list,
          reward       = quest_data.reward,
          target       = quest_data.target,
          section      = section_name
        } );

      end;

    end;

  end;
  
end;
--******************************************************************************
function private.GetPregiveQuests ( quest_type, section_name, source, quests_current, quests_out )
  
  -- AL> Квесты с pregive-механикой
  if ( source == "stage" ) then

    -- Get all pregive quests and check 'current' and 'completed' state
    local query =
    {
      query      = "find_all",
      group      = "static",
      collection = "quests",
      indice     = "target.pregive",
      criteria   = { like = "index" }
    };
    
    local quests_pregive = m_storage.Execute( query ).result;

    for _, quest in pairs( quests_pregive ) do

      -- Check if pregive quest is current
      local quest_is_current = false;

      for i = 1, #quests_current do

        if ( quest.name == quests_current[ i ].name ) then

          quest_is_current = true;
          break;

        end;

      end;
      
      -- Chek if pregive quest is completed
      local quest_is_complete = false;

      local current_query =
      {
        query      = "find_all",
        group      = "player",
        collection = "quests_completed"
      };
      
      local quests_completed = m_storage.Execute( current_query ).result;

      for i = 1, #quests_completed do

        if ( quest.name == quests_completed[ i ].name ) then

          quest_is_complete = true;
          break;

        end;

      end;
      
      -- If qest not current and not completed, add them to quests_out table
      if ( not quest_is_current ) and ( not quest_is_complete ) then

        private.TryStartQuests( quest.condition, true );

        if ( quest_type == quest.target.type ) then

          if ( quest.reward ) then

            if ( #quest.reward == 0 ) then -- FIXME: WTF?!

              local rewards = {};
              table.insert( rewards, quest.reward );
              quest.reward = rewards;

            end;

            for r = 1, #quest.reward do

              quest.reward[ r ].content = 
                m_content.GetContent( quest.reward[ r ].type, 
                                      quest.reward[ r ].name );

            end;

          end;

          table.insert( quests_out,
          {
            name    = quest.name,
            pregive = true,
            count   = quest.count,
            list    = quest.list,
            reward  = quest.reward,
            target  = quest.target,
            section = section_name
          } );

        end;

      end;

    end;

  end;

end;
--******************************************************************************
function private.GetLinearQuests ( 
  quest_type, 
  section_name, 
  source, 
  quests_out 
)
  
  -- AL> Линейно-генерируемые квесты
  if 
    ( ( not quest_type ) or ( quest_type == "QT_StageCompleted" or
    quest_type == "QT_BuildConstructed" ) ) and 
    ( source == "external" ) then

    local stages_info  = m_stage.GetStagesStat();
    local section_num  = tonumber( section_name:match( "%d+" ) );
    local previous_count = 0;

    if ( section_num > 1 ) then
      for i = 1, section_num - 1 do
        local section_data = stages_info[ "section_"..section_num ];
        previous_count = 
          previous_count + 
          section_data.total_count - 
          section_data.bonus_total;
      end;
    end;

    local stages_count = stages_info[ section_name ].completed_count;
    local stages_total = 
      stages_info[ section_name ].total_count - 
      stages_info[ section_name ].bonus_total;
    local target_count = previous_count + stages_count + 1;
    local target_name  = 
      private.default.complete_stage_quest_name .. target_count;
    local stage_name   = "stage_"..target_count;
    local custom_quest = public.GetQuestData( target_name );
    
    if ( not custom_quest ) then

      local saved_stage_data = m_stage.GetSavedData( stage_name );
      
      if ( stages_count < stages_total ) and 
         ( saved_stage_data and saved_stage_data.visible ) then
        
        if ( private.caches.current_quests_stage ~= stage_name ) then
          
          mediator.Notify( "Model_Quests_QuestStarted", target_name );
          
        end;
        
        private.caches.current_quests_stage = stage_name;
        
        table.insert( quests_out,
        {
          name   = target_name,
          hidden = false,
          target = {
            type = "QT_StageCompleted",
            name = stage_name
          },
          section = section_name
        } );
        
      end;

    else

      private.caches.current_quests_stage = nil;

    end;

  end;
  
end;
  
function private.GetDynamicQuests ( 
  quest_type, 
  section_name, 
  source, 
  quests_out 
)

  -- AL> Динамически-генерируемые квесты 
  if 
  ( not quest_type or quest_type == "QT_MineResourceCollected" ) and 
  ( private.ready_mines_list[ section_name ] ) and 
  ( #private.ready_mines_list[ section_name ] > 0 ) then

    for m = 1, #private.ready_mines_list[ section_name ] do

      local mine_name = private.ready_mines_list[ section_name ][ m ];

      table.insert( quests_out,
      {
        name   = private.default.mine_collect_quest_name..mine_name,
        target = {
          type = "QT_MineResourceCollected",
          name = mine_name
        },
        section = section_name
      } );

    end;

  end;

  if 
  ( not quest_type or quest_type == "QT_SpellCompleted" ) and 
  ( private.enabled_spells_list[ section_name ] ) and
  ( #private.enabled_spells_list[ section_name ] > 0 ) then

    for s = 1, #private.enabled_spells_list[ section_name ] do

      local spell_name = private.enabled_spells_list[ section_name ][ s ];

      table.insert( quests_out,
      {
        name   = private.default.complete_spell_quest_name .. spell_name,
        target = {
          type = "QT_SpellCompleted",
          name = spell_name
        },
        section = section_name
      } );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.IsStageItemActive ( item_name, stage_name, source, section )

  local quest_list = private.GetCurrentQuestsInfo( "QT_StageItemsCollected", source, section );

  if ( #quest_list > 0 ) then

    for q = 1, #quest_list do

      if ( quest_list[ q ].target.name == item_name ) then

        local target_list  = quest_list[ q ].target.list;
        local current_list = quest_list[ q ].list;
        local pregive_list = quest_list[ q ].target.pregive;
        
        if ( quest_list[ q ].pregive ) then

          if (  cmn.IsEntityInList( pregive_list, stage_name )) and
          ( not cmn.IsEntityInList( current_list, stage_name )) then

            return true;

          end;

        else

          if (  cmn.IsEntityInList( target_list,  stage_name )) and
          ( not cmn.IsEntityInList( current_list, stage_name )) then

            return true;

          end;

        end;

      end;

    end;

  end;

  return false;

end;
--******************************************************************************
--function ***** EVENTS ***** ()
--******************************************************************************
function private.EventParser ( event_type, params )

  -- AL> Обновление активных квестов
  local quests_current = private.GetCurrentQuestsInfo( "QT_"..event_type, "event_parser" );

  if ( #quests_current > 0 ) then

    if     ( event_type == "StageCompleted" ) then

      private.UpdateType_StageCompleted( quests_current, params );

    elseif ( event_type == "StageListCompleted" ) then
    
      private.UpdateType_StageListCompleted( quests_current, params );

    elseif ( event_type == "ReachMaxLevelStages" ) then
    
      private.UpdateType_ReachMaxLevelStages( quests_current, params );

    elseif ( event_type == "StageItemsCollected" ) then
    
      private.UpdateType_StageItemsCollected( quests_current, params );

    elseif ( event_type == "FragmentCollected" ) then
    
      private.UpdateType_FragmentCollected( quests_current, params );

    elseif ( event_type == "SpellCompleted" ) then

      private.UpdateType_SpellCompleted( quests_current, params );

    elseif ( event_type == "BuildConstructed" ) then
  
      private.UpdateType_BuildConstructed( quests_current, params );

    elseif ( event_type == "MineConstructed" ) then

      private.UpdateType_MineConstructed( quests_current, params );

    elseif ( event_type == "MineResourceCollected" ) then

      private.UpdateType_MineResourceCollected( quests_current, params );

    elseif ( event_type == "MarketConstructed" ) then

      private.UpdateType_MarketConstructed( quests_current, params );

    end;

  elseif ( event_type == "StageCompleted" ) and ( private.caches.current_quests_stage == params.name )then

    private.UpdateType_StageCompleted( quests_current, params );

  end;

  -- AL> Проверка квестов на возможность старта
  if ( event_type == "MineResourceFilled" ) then

    private.StartType_MineResourceCollected( params );

  elseif ( event_type == "SpellEnabled" ) then

    private.StartType_SpellEnabled( params );

  elseif
  ( event_type == "QuestCompleted" ) or
  ( event_type == "StageCompleted" ) or
  ( event_type == "SpellCompleted" ) or
  ( event_type == "BuildConstructed" ) or
  ( event_type == "MineConstructed" ) or
  ( event_type == "MineResourceCollected" ) or
  ( event_type == "MineResourceFilled" ) then

    local quest_condition = { 
      type  = "QSC_"..event_type, 
      name  = params.name
    };
    private.TryStartQuests( quest_condition, false );

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_StageCompleted ( quests_current, params )

  if ( not params.enable_next ) then

    return;

  end;

  local stage_name = params.name;
  
  if ( private.caches.current_quests_stage == stage_name ) then
    
    private.caches.current_expired = true;

    mediator.Notify( "Model_Quests_QuestCompleted", {
      name    = "qst_complete_"..private.caches.current_quests_stage,
      section = private.current_section,
      target  = {
        type  = "QT_StageCompleted"
      }
    } );
              
    return;

  end;

  for q = 1, #quests_current do

    local quest = quests_current[ q ];

    if ( not quest.target.name ) or ( stage_name == quest.target.name ) then

      private.CompleteQuest( quest );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_StageListCompleted ( quests_current, params )

  local stage_name = params.name;

  for q = 1, #quests_current do

    local quest = quests_current[ q ];

    if ( cmn.IsEntityInList( quest.target.list, stage_name ) ) then

      private.UpdateQuestList( quest, stage_name );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_ReachMaxLevelStages ( quests_current, params )

  local stage_name = params.name;
  local state_new  = params.state_new;
  local state_old  = params.state_old;

  for q = 1, #quests_current do

    local quest = quests_current[ q ];
    local target_list = quest.target.list;

    for t = 1, #target_list do

      if ( stage_name == target_list[ t ] ) and ( state_old == state_new ) and ( state_new == 3 ) then

        private.UpdateQuestList( quest, stage_name );
        break;

      end;

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_StageItemsCollected ( quests_current, params )

  local stage_name = params.stage;
  local item_name  = params.name;

  for q = 1, #quests_current do

    local quest = quests_current[ q ];

    if ( quest.target.name == item_name ) then

      if ( cmn.IsEntityInList( quest.target.list, stage_name ) ) then

        private.UpdateQuestList( quest, stage_name );

      end;

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_FragmentCollected ( quests_current, params )

  local item_name  = params.name;

  for q = 1, #quests_current do

    local quest = quests_current[ q ];

    if ( cmn.IsEntityInList( quest.target.list, item_name ) ) then

      private.UpdateQuestList( quest, item_name );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_BuildConstructed ( quests_current, params )

  local build_name = params.name;

  for q = 1, #quests_current do

    local quest = quests_current[ q ];
    local target_name = quest.target.name;

    if ( not target_name ) or ( build_name == target_name ) then

      private.CompleteQuest( quest );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_MineConstructed ( quests_current, params )

  local mine_name = params.name;

  for q = 1, #quests_current do

    local quest = quests_current[ q ];
    local target_name = quest.target.name;

    if ( not target_name ) or ( mine_name == target_name ) then

      private.CompleteQuest( quest );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_MarketConstructed ( quests_current, params )

  for q = 1, #quests_current do

    private.CompleteQuest( quests_current[ q ] );

  end;

end;
--******************************************************************************
-- function ***** EVENTS - GENERATED ***** ()
--******************************************************************************
function private.GetReadyMines ( quests_current, params )

  local result = {};

  local query =
  {
    query      = "find_all",
    group      = "player",
    collection = "mines",
    indice     = "state",
    criteria   = { like = "ready" }
  };

  local mine_list = m_storage.Execute( query ).result;

  if ( mine_list ) and ( #mine_list > 0 ) then

    for m = 1, #mine_list do

      local mine_name    = mine_list[ m ].name;
      local mine_section = private.GetMineSection( mine_name );

      if ( not result[ mine_section ] ) then
        result[ mine_section ] = {};
      end;

      table.insert( result[ mine_section ], mine_name );

    end;

  end;

  return result;

end;
--******************************************************************************
function private.GetMineSection ( mine_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "mines",
    indice     = "name",
    criteria   = { like = mine_name }
  };

  return m_storage.Execute( query ).result.section;

end;
--******************************************************************************
--******************************************************************************
function private.StartType_MineResourceCollected ( params )

  local mine_name = params.name;
  local mine_section = private.GetMineSection( mine_name );

  if ( cmn.IsEntityInList( private.ready_mines_list[ mine_section ], mine_name )) then

    return;

  end;

  if not ( private.ready_mines_list[ mine_section ] ) then

    private.ready_mines_list[ mine_section ] = {};

  end;

  table.insert( private.ready_mines_list[ mine_section ], mine_name );

  private.caches.current_expired = true;

  local quest_name = private.default.mine_collect_quest_name .. mine_name;

  mediator.Notify( "Model_Quests_QuestStarted", quest_name );

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_MineResourceCollected ( quests_current, params )

  if ( params.start_full ) then
    return;
  end;
  
  local mine_name = params.name;

  for q = 1, #quests_current do

    local quest_name  = quests_current[ q ].name;
    local target_name = quests_current[ q ].target.name;

    if ( not target_name ) or ( mine_name == target_name ) then

      local mine_section = private.GetMineSection( mine_name );

      for m = 1, #private.ready_mines_list[ mine_section ] do

        if ( private.ready_mines_list[ mine_section ][ m ] == target_name ) then

          table.remove( private.ready_mines_list[ mine_section ], m );

          break;

        end;

      end;

      private.caches.current_expired = true;

      private.EventParser( "QuestCompleted", { name = quest_name } );

      mediator.Notify( "Model_Quests_QuestCompleted", { 
        name    = quest_name,
        section = mine_section,
        target  = {
          type  = quests_current[ q ].target.type
        }
      } );
      
    end;

  end;

end;
--------------------------------------------------------------------------------
function private.GetEnabledSpells ( quests_current, params )

  local result = {};

  local query = 
  {
    query      = "find_all",
    group      = "player",
    collection = "spells"
  };
  
  local spell_data = m_storage.Execute( query ).result;

  for s = 1, #spell_data do

    local spell_name = spell_data[ s ].name;

    if ( spell_data[ s ].enabled ) and ( not spell_data[ s ].completed ) then

      local spell_section = private.GetSpellSection( spell_name );

      if ( not result[ spell_section ] ) then
        result[ spell_section ] = {};
      end;

      table.insert( result[ spell_section ], spell_name );

    end;

  end;

  return result;

end;
--------------------------------------------------------------------------------
function private.GetSpellSection ( spell_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "spells",
    indice     = "name",
    criteria   = { like = spell_name }
  };

  return m_storage.Execute( query ).result.section;

end;
--******************************************************************************
--******************************************************************************
function private.StartType_SpellEnabled ( params )

  local spell_name = params;
  local spell_section = private.GetSpellSection( spell_name );

  if ( cmn.IsEntityInList( private.enabled_spells_list[ spell_section ], spell_name )) then

    return;

  end;

  if not ( private.enabled_spells_list[ spell_section ] ) then

    private.enabled_spells_list[ spell_section ] = {};

  end;

  table.insert( private.enabled_spells_list[ spell_section ], spell_name );

  private.caches.current_expired = true;

  local quest_name = private.default.complete_spell_quest_name .. spell_name;

  mediator.Notify( "Model_Quests_QuestStarted", quest_name );

end;
--******************************************************************************
--******************************************************************************
function private.UpdateType_SpellCompleted ( quests_current, params )

  local spell_name = params;

  for q = 1, #quests_current do

    local quest_name  = quests_current[ q ].name;
    local target_name = quests_current[ q ].target.name;

    if ( not target_name ) or ( spell_name == target_name ) then

      local spell_section = private.GetSpellSection( spell_name );

      for s = 1, #private.enabled_spells_list[ spell_section ] do

        if ( private.enabled_spells_list[ spell_section ][ s ] == target_name ) then

          table.remove( private.enabled_spells_list[ spell_section ], s );

          break;

        end;

      end;

      private.caches.current_expired = true;

      private.EventParser( "QuestCompleted", { name = quest_name } );

      mediator.Notify( "Model_Quests_QuestCompleted", { 
        name    = quest_name,
        section = spell_section,
        target  = {
          type  = quests_current[ q ].target.type
        }
      } );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.UpdateQuestList ( quest, current_element )

  local quest_name   = quest.name;
  local target_list  = quest.target.list;
  local current_list = quest.list;

  if ( cmn.IsEntityInList( current_list, current_element )) then

    return;

  end;

  local completed = false;

  local new_list = {};

  if ( current_list ) then

    new_list = current_list;

  end;

  table.insert( new_list, current_element );

  local update_set = { list = new_list };

  local query = 
  {
    query      = "update",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name },
    set        = update_set
  };
  
  m_storage.Execute( query );

  private.caches.current_expired = true;

  if ( #new_list >= #target_list ) then

    completed = true;

  end;

  mediator.Notify( "Model_Quests_QuestUpdated", quest_name );

  if ( completed ) then

    private.CompleteQuest( quest );

  end;

end;
--------------------------------------------------------------------------------
function private.Quest_Progress ( quest_name )

  local query = 
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = {
      like = quest_name
    }
  };
  
  return m_storage.Execute( query ).result.list;

end;
--******************************************************************************
--******************************************************************************
function private.CompleteQuest ( quest )

  local quest_name = quest.name;

  local query = 
  {
    query      = "delete",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  m_storage.Execute( query );

  query = 
  {
    query      = "insert",
    group      = "player",
    collection = "quests_completed",
    value      = { name = quest_name }
  };
  
  m_storage.Execute( query );

  private.caches.current_expired = true;

  private.EventParser( "QuestCompleted", { name = quest_name } );

  mediator.Notify( "Model_Quests_QuestCompleted", { 
    name = quest_name, 
    section = quest.section ,
    target  = {
      type  = quest.target.type
    }
  } );

  if ( quest.reward ) then

    for r = 1, #quest.reward do

      quest.reward[ r ].category = "quest";
      quest.reward[ r ].quest_name = quest_name;
      
      m_items.TryObtain( quest.reward[ r ] );

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.TryStartQuests ( quest_condition, pregive )

  local func_get_save_info = function ( quest_name, group, collection, indice )

    local query =
    {
      query      = "find",
      group      = group,
      collection = collection,
      indice     = indice,
      criteria   = { like = quest_name }
    };
    
    return m_storage.Execute( query ).result;

  end;

  -- AL> parts:db: lua.table не может быть критерием поиска
  local started_query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "condition.type",
    criteria   = { like = quest_condition.type }
  };
  
  local quests_candidates = m_storage.Execute( started_query ).result;

  for c, candidate in ipairs( quests_candidates ) do

    if 
    ( candidate.condition.name and ( candidate.condition.name == quest_condition.name )) or
    ( candidate.condition.list and ( cmn.IsEntityInList( candidate.condition.list, quest_condition.name ))) then

    local quest_current_info  = func_get_save_info( candidate.name, "player", "quests_current",   "name" );
    local quest_complete_info = func_get_save_info( candidate.name, "player", "quests_completed", "name" );

      if ( not pregive ) and ( quest_current_info ) and ( quest_current_info.pregive ) then

        local update_set = quest_current_info;
        update_set.pregive = false;

        local update_query =
        {
          query      = "update",
          group      = "player",
          collection = "quests_current",
          indice     = "name",
          criteria   = { like = candidate.name },
          set        = update_set
        };
        
        m_storage.Execute( update_query );

        private.caches.current_expired = true;

        mediator.Notify( "Model_Quests_QuestStarted", candidate.name );

      elseif ( not quest_current_info ) and ( not quest_complete_info ) then

        local insert_value = private.CreateQuestInitValue( candidate );

        if ( pregive ) then

          insert_value.pregive = true;

        end;

        local insert_query =
        {
          query      = "insert",
          group      = "player",
          collection = "quests_current",
          value      = insert_value
        };
        
        m_storage.Execute( insert_query );

        private.caches.current_expired = true;

        if ( not pregive ) then

          mediator.Notify( "Model_Quests_QuestStarted", candidate.name );

        else

          mediator.Notify( "Model_Quests_PregiveStarted", candidate.name );

        end;

      end;

    end;

  end;

end;
--******************************************************************************
--******************************************************************************
function private.CreateQuestInitValue ( started_quest )

  local quest_name = started_quest.name;
  local quest_type = started_quest.target.type;

  local insert_value = { name = quest_name };

  if     ( quest_type == "QT_StageCompleted" ) then

    insert_value.count = 0;

  elseif ( quest_type == "QT_ReachMaxLevelStages" ) then

    local insert_list = {};
    local target_list = started_quest.target.list;

    for t = 1, #target_list do 

      local stage_name  = target_list[ t ];
      local stage_level = m_stage.GetStageState( target_list[ t ] );

      local query =
      {
        query      = "find_all",
        group      = "player",
        collection = "stages"
      };
      
      local save_stage_list = m_storage.Execute( query ).result;

      local func_can_add = function ( stage_list, checked_stage_name )

        for i = 1, #stage_list do

          if ( stage_list[ i ].name == checked_stage_name ) then

            if ( stage_list[ i ].state ) and ( stage_list[ i ].state >= 3 ) and ( stage_list[ i ].noreward ) then

              return true;

            else

              return false;

            end;

          end;

        end;

        return false;

      end;

      if ( func_can_add( save_stage_list, stage_name ) ) then

        table.insert( insert_list, stage_name );

      end;

    end;

    if ( #insert_list > 0 ) then

      insert_value.list = insert_list;

    end;

  elseif ( quest_type == "QT_FragmentCollected" ) then

    local insert_list = {};
    local target_list = started_quest.target.list;
    local build_name  = m_builds.GetBuildNameByFragment( target_list[ 1 ] );
    local fragments_collected = m_builds.GetBuildData( build_name ).collected_fragments;

    for t = 1, #target_list do 

      if ( cmn.IsEntityInList( fragments_collected, target_list[ t ] )) then

        table.insert( insert_list, target_list[ t ] );

      end;

    end;

    if ( #insert_list > 0 ) then

      insert_value.list = insert_list;

    end;

  --[[ AL> parts:db: нельзя писать пустые массивы т.к. невозможно вангануть их тип
  elseif ( quest_type == "QT_StageListCompleted" ) then

    insert_value.list = {};

  elseif ( quest_type == "QT_StageItemsCollected" ) then

    insert_value.list = {};
  ]]

  end;

  return insert_value;

end;
--******************************************************************************
-- function ***** SERVICE ***** ()
--******************************************************************************
function public.GetQuestData ( quest_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  return m_storage.Execute( query ).result;
  
end;
--******************************************************************************
--******************************************************************************
function private.Name_ByTarget ( target_name )

  local query =
  {
    query      = "find",
    group      = "static",
    collection = "quests",
    indice     = "target.name",
    criteria   = { like = target_name }
  };

  return m_storage.Execute( query ).result.name;

end;
--******************************************************************************
--******************************************************************************
function public.CheaterUpdate( stage_name, item_name )

  local quest_name = private.Name_ByTarget( item_name );
  local quest_data = public.GetQuestData( quest_name );
  
  local query =
  {
    query      = "find",
    group      = "player",
    collection = "quests_current",
    indice     = "name",
    criteria   = { like = quest_name }
  };
  
  local quest_current = m_storage.Execute( query ).result;

  if ( quest_current ) then

    if ( quest_current.list ) then

      quest_data.list = quest_current.list;

    end;

    private.UpdateQuestList( quest_data, stage_name );

  else

    local insert_data = {
      name = quest_name,
      list = { stage_name }

    };

    if ( quest_data.pregive ) then

      insert_data.pregive = true;

    end;

    local query =
    {
      query      = "insert",
      group      = "player",
      collection = "quests_current",
      value      = insert_data
    };

    m_storage.Execute( query );

  end;

end;
--------------------------------------------------------------------------------
function public.GetQuestsOnStageComplete( stage_name )

  local result = {};
  
  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "condition.type",
    criteria   = { like = "QSC_StageCompleted" }
  };

  local list = m_storage.Execute( query ).result;
  
  for i, quest in pairs( list ) do
  
    if ( quest.condition.name == stage_name ) then
    
      table.insert( result, quest );
    
    end;
  
  end;
  
  return result;

end;
--------------------------------------------------------------------------------
function public.IsItemPregived ( item )

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "target.pregive",
    criteria   = { like = "index" }
  };
    
  local quests = m_storage.Execute( query ).result or {};
  
  for _, quest in pairs( quests ) do
  
    if ( item.name == quest.target.name ) then
    
      return true;
    
    end;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function public.StageList_HasReward (
    section_name,
    reward_type,
    reward_name_pattern
  )

  local query =
  {
    query      = "find_all",
    group      = "static",
    collection = "quests",
    indice     = "section",
    criteria   = {
      like = section_name
    }
  };
    
  local quests = m_storage.Execute( query ).result or {};

  if ( #quests < 1 ) then

    return quests;

  end;

  local list = {};

  for q, quest in pairs( quests ) do

    if ( quest.reward                             ) and
       ( not public.IsQuestComplete( quest.name ) ) and
       (     public.IsQuestCurrent( quest.name )  ) then

      local quest_progress = private.Quest_Progress( quest.name );

      for r, reward in pairs( quest.reward ) do

        if ( reward.type == reward_type              ) and
           ( reward.name:find( reward_name_pattern ) ) then

          for s, stage_name in pairs( quest.target.list ) do

            if ( not cmn.IsEntityInList( quest_progress, stage_name ) ) then

              table.insert( list, stage_name );

            end;

          end;

        end;

      end;

    end;

  end;

  return list;

end;
--------------------------------------------------------------------------------