-- name=v_characters
--------------------------------------------------------------------------------
--[[

  Character model states:
    + locked   -- character not available
    + unlocked -- character available, but has no quests started
    + active   -- character available and has active quest
    + inactive -- character available and all quests done

  Character visual states:
    + normal -- character in normal state
    + cage   -- character in caged in cell
    + charm  -- charcter charmed to ghost

    * 'cage' and 'charm' visual states has same named animation resources,
      but character behavior and animfuncs are different

]]
--------------------------------------------------------------------------------
function public.__InitModule ()

  loader.Module_Load( "int_dialog_map_quest_reward" );

  -- Stateless done
  private.show_states =
  {
    quest  = "qst_collect_star",
    stages = {}
  };
  
  for s = 1, m_stage.GetStatelessNum() - 1 do
    table.insert( private.show_states.stages, "stage_"..s );
  end;

end;
--------------------------------------------------------------------------------
function public.Create ( character, section )

  local obj_name = ( "obj_map_characters_%s" ):format( character );
  local btn_name = ( "btn_map_characters_%s" ):format( character );
  
  -- ObjCreate( obj_name, "obj"                    );
  ObjSet   ( obj_name, { anim_tag = character } );
  ObjAttach( obj_name, "anm_map_"..section      );
  
  ObjClone( "btn_map_characters", "_"..character, 1, obj_name );
  
  local params = ObjGet( obj_name );
  
  ObjSet( obj_name, { inputrect_init = false } );
  ObjSet( btn_name,
    {
      inputrect_init = true,
      inputrect_x = params.inputrect_x,
      inputrect_y = params.inputrect_y,
      inputrect_w = params.inputrect_w,
      inputrect_h = params.inputrect_h
    }
  );
  
  local func_click = function () public.Tap( character ) end;
  int.Control_EventsSet( btn_name, { func_on_click = func_click } );
  
  local state = m_characters.GetCharacterState( character );

  if     ( state == "locked"   ) then private.SetLocked( character );
  elseif ( state == "unlocked" ) then private.SetUnlocked( character );
  elseif ( state == "active"   ) then private.SetActive( character );
  elseif ( state == "inactive" ) then private.SetInactive( character );
  end;

end;
--------------------------------------------------------------------------------
function private.SetLocked ( character )

  m_object.Disable( "btn_map_characters_"..character );

end;
--------------------------------------------------------------------------------
function private.SetUnlocked ( character )

  m_object.Enable( "btn_map_characters_"..character );
  private.Indicator_Disable( character, false );

  local state = private.GetVisualState_Unlocked( character );
  private.SetVisualState( character, state );
  
end;
--------------------------------------------------------------------------------
function private.SetActive ( character )
  
  m_object.Enable( "btn_map_characters_"..character );
  private.Indicator_Enable( character, true );
  
  local quest = m_characters.GetCurrentQuest( character );
  
  if ( quest ) then
    local state = private.GetVisualState_Active( character, quest.name );
    private.SetVisualState( character, state );
  else
    DbgTrace( "[ ERROR ] SetActive: no quest for character < "..character.." >!" );
    private.SetVisualState( character, "normal" );
  end;
  
end;
--------------------------------------------------------------------------------
function private.SetInactive ( character )
  
  m_object.Enable( "btn_map_characters_"..character );
  private.Indicator_Disable( character );
  
  private.SetVisualState( character, "normal" );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function public.Tap ( character )

  local m_state = m_characters.GetCharacterState( character );
  local v_state = "normal";
  
  if     ( m_state == "unlocked" ) then
  
    v_state = private.GetVisualState_Unlocked( character );
    private.Popup_Unlocked( character );
  
  elseif ( m_state == "active"   ) then
  
    local quest = m_characters.GetCurrentQuest( character );
    if ( quest ) then
      quest.content = m_content.GetContent( "quest", quest.target.name );
      v_state = private.GetVisualState_Active( character, quest.name );
      private.Popup_Active( character, quest );
    else
      DbgTrace( "[ ERROR ] Tap: no quest for character < "..character.." >!" );
    end;
  
  elseif ( m_state == "inactive" ) then
  
    private.Popup_Inactive( character );
  
  elseif ( IsEditor()          ) then
  
    local msg = "Error! Character %q in wrong state %q";
    common.DbgCallpoint( msg:format( character, m_state ) );
  
  end;
  
  private.Animate_Action( character, v_state );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Indicator_Enable ( character, show )

  local ind_name = "anm_map_characters_quest_"..character;
  m_object.Enable( ind_name, true  );
  
  if ( show ) then
  
    private.Indicator_Show( character );
  
  end

end;
--------------------------------------------------------------------------------
function private.Indicator_Disable ( character, hide )

  local ind_name = "anm_map_characters_quest_"..character;
  
  if ( hide ) then
  
    private.Indicator_Hide( character, function ()
    
      m_object.Disable( ind_name, true );
    
    end );
  
  else
  
    m_object.Disable( ind_name, true );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Indicator_Show ( character )

  local ind_name = "anm_map_characters_quest_"..character;
  AnimPlay( ind_name, "quest_popup_open", "" );

end;
--------------------------------------------------------------------------------
function private.Indicator_Hide ( character, func_after )

  local ind_name = "anm_map_characters_quest_"..character;
  AnimPlay( ind_name, "quest_popup_close", func_after or "" ); 

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Sound_Play ( sound_id )
  
  cmn.PlaySnd( "assets/content/map/sounds/aud_"..sound_id );
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
--[[
  Character reaction format:
  {
    -- Basic data:
    name       = "character_name"
    func_popup = function ( func_after ) end,
    func_anim  = function ( func_after ) end,
    
    -- Only for reward drop:
    object          = "obj_char_name", -- receiver-object
    func_drop       = function () end, -- on-drop function
    call_drop_first = true, -- true : func_drop on focus, func_anim on drop
                            -- false: func_anim on focus, func_drop on drop
    
    -- Rewards:
    quest_done    = "completed_quest_name",
    reward        = { ... }, -- quest rewards table
    func_reward   = function ( reactions, func_after ) end,
    skinpart      = "skn_part_name",
    story_data    = { ... }, -- See m_story.GetData
    
    -- Other:
    update_stages = { "stage_1", "stage_2", "stage_3", ... }
  }
]]
--------------------------------------------------------------------------------
function public.GetRewardReaction ( stage, rewards )

  local reaction = nil;
  
  if not ( rewards ) then return reaction; end;
  
  -- Find quest item
  local quest_item = nil;
  for i, reward in pairs( rewards ) do
  
    if ( reward.type == "quest" ) then
    
      quest_item = reward;
      break;
    
    end;
  
  end;
  
  if not ( quest_item ) then return reaction; end;
  
  local item_name = quest_item.name;
  
  -- Find character and quest
  local character  =
    m_characters.GetCharacterByItemFromStage( item_name, stage );
  
  local quest      =
    m_characters.GetQuestByItemFromStage( item_name, stage );
  
  quest.content = m_content.GetContent( "quest", item_name );
  
  local quest_done = 
    m_characters.IsQuestForItemFromStageDone( item_name, stage );
  
  -- Setup reaction
  reaction        = {};
  reaction.name   = character;
  reaction.object = "obj_map_characters_"..character;
  
  reaction.call_drop_first = false;
  reaction.func_drop       = nil;
  
  reaction.func_popup = function ( func_after )
    private.Popup_ItemReceive( character, quest, func_after );
  end;
  
  if ( quest_done ) then
  
    reaction.quest_done = quest.name;
    reaction.reward = quest.reward;
    
    for i, reward in pairs( quest.reward ) do
    
      if ( reward.type == "skinpart" ) then
      
        reaction.skinpart = reward.name;
        break;
      
      end;
    
    end;
    
    local section   = m_map.Section_Current_Name();
    local condition = "CompleteQuest";
    reaction.story_data = m_story.GetData( section, condition, quest.name );
    
    local animation    = private.GetAnimData_QuestDone( character, quest );
    reaction.func_anim = v_map.Callback_Animate( animation );
    
    reaction.func_reward = function ( reactions, func_after )
      private.ShowReward( character, quest.reward, reactions, func_after );
    end;
    
  else
  
    local animation    = private.GetAnimData_ItemReceive( character, quest );
    reaction.func_anim = v_map.Callback_Animate( animation );
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function public.GetQuestCompleteReaction ( quest_name )

  local reaction = nil;
  if not ( quest_name ) then return reaction; end;

  local quest_next = nil;
  local character = m_characters.GetCharacterByQuest( quest_name );
  
  if ( character ) then
  
    if ( quest_name == character.quest_cage ) then
      quest_next = character.quest_free;
    elseif ( quest_name == character.quest_free ) then
      quest_next = character.quest_cage;
    end;
  
  end;
  
  if ( quest_next and m_quests.IsQuestCurrent( quest_next ) ) then
  
    local quest = m_quests.GetQuestData( quest_next );
    quest.content = m_content.GetContent( "quest", quest.target.name );
  
    -- Setup reaction
    reaction               = {};
    reaction.name          = character.name;
    reaction.object        = "obj_map_characters_"..character.name;
    
    reaction.func_popup    = function ( func_after )
    
      private.Popup_Active( character.name, quest, func_after );
      
    end;
    
    local animation = private.GetAnimData_QuestStart( character.name, quest );
    reaction.func_anim = v_map.Callback_Animate( animation );
    
    -- HACK for finish staeless mode
    if ( quest.name == private.show_states.quest ) then
    
      reaction.update_stages = private.show_states.stages;
    
    else
    
      reaction.update_stages = quest.target.list;
    
    end;
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function public.GetStageCompleteReaction ( stage )

  local reaction = nil;
  local character, quest;
  
  local quests = m_quests.GetQuestsOnStageComplete( stage );
  for _, stage_quest in pairs( quests ) do
  
    character = m_characters.GetCharacterByQuest( stage_quest.name );
    if ( character ) then
    
      quest = stage_quest;
      break;
    
    end;
  
  end;
  
  if ( character ) then
  
    quest.content = m_content.GetContent( "quest", quest.target.name );
  
    local name = character.name;
  
    reaction               = {};
    reaction.name          = name;
    reaction.object        = "obj_map_characters_"..name;
    reaction.update_stages = quest.target.list;
    reaction.func_popup    = function ( func_after )
    
      private.Popup_Active( name, quest, func_after );
      
    end;
    
    local animation    = private.GetAnimData_QuestStart( name, quest );
    reaction.func_anim = v_map.Callback_Animate( animation );
    
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function private.ShowReward ( character, rewards, reactions, func_after )

  local func_close = function ()

    int_dialog_map_quest_reward.Hide(
      function () 
        v_map.Reward_Show( false, rewards, reactions, func_after );
      end, 
      function ()
        int_window.Hide( "int_dialog_map_quest_reward" );
      end
    );

  end;

  local func_start = function ()

    int.Lock( false );
    v_map.Interface_Disable( true );

    local content = m_characters.GetContent( character );

    int_dialog_map_quest_reward.Show(
      character,
      content,
      rewards,
      func_close
    );

  end;

  int_window.Show(
    "int_dialog_map_quest_reward",
    func_start, 
    { 
      map_active = true
    }
  );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Popup_Unlocked ( character, func_after )
  
  v_map.ShowObjectPopup(
    "monolog",
    "obj_map_characters_indicator_"..character,
    "str_int_popup_cloud_characters_"..character,
    nil,
    nil,
    {
      func_preclose = func_after
    }
  );
  
end;
--------------------------------------------------------------------------------
function private.Popup_ItemReceive ( character, quest, func_after )

  private.Indicator_Hide( character );

  local func_close = function ()
    private.Indicator_Show( character );
  end;
  
  local data = m_characters.GetQuestPopupData( quest );

  v_map.ShowObjectPopup(
    "counter",
    "obj_map_characters_indicator_"..character,
    "",
    data,
    nil,
    {
      func_preclose = func_after,
      func_close    = func_close
    }
  );

end;
--------------------------------------------------------------------------------
function private.Popup_Active ( character, quest, func_after )

  private.Indicator_Hide( character );

  local func_close = function ()
    private.Indicator_Show( character );
  end;
  
  local data = m_characters.GetQuestPopupData( quest );
  
  v_map.ShowObjectPopup(
    "complex",
    "obj_map_characters_indicator_"..character,
    "str_int_popup_cloud_characters_"..character,
    data,
    nil,
    {
      func_preclose = func_after,
      func_close    = func_close
    }
  );
  
end;
--------------------------------------------------------------------------------
function private.Popup_Inactive ( character, func_after )
  
  local popup = "str_int_popup_cloud_characters_"..character;
  
  v_map.ShowObjectPopup(
    "monolog",
    "obj_map_characters_indicator_"..character,
    "str_int_popup_cloud_characters_completed_"..character,
    nil,
    nil,
    {
      func_preclose = func_after
    }
  );
  
end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.GetVisualState_Unlocked ( character )

  local data       = m_characters.GetCharacterData( character );
  local state_type = data.unlocked_type;
  
  local states =
  {
    cage = "cage",
    free = "normal"
  };
  
  return states[ state_type ];

end;
--------------------------------------------------------------------------------
function private.GetVisualState_Active ( character, quest_name )

  local data = m_characters.GetCharacterData( character );
  
  local quest_cage = data.quest_cage;
  local quest_free = data.quest_free;
  local state_type = data.unlocked_type;
  
  -- DG: Characters data format workaround
  -- Two quests + unlock type free = reverted character states
  if ( quest_cage and quest_free and state_type == "free" ) then
  
    local states =
    {
      [ quest_cage ] = "normal",
      [ quest_free ] = "charm"
    }
    
    return states[ quest_name ];
  
  end;
  
  local states = {};
  
  if ( quest_cage ) then states[ quest_cage ] = "cage";   end;
  if ( quest_free ) then states[ quest_free ] = "normal"; end;
  
  return states[ quest_name ];

end;
--------------------------------------------------------------------------------
function private.SetVisualState ( character, state )

  local anm_name = "anm_map_characters_icon_"..character;
  local content = m_characters.GetContent( character );

  local normal =
  {
    res = "assets/content/"..content.."/anm_character_"..character,
    animfunc = "static",
    playing = true
  }
  
  local cell =
  {
    res = "assets/content/"..content.."/anm_character_cell",
    animfunc = "static_"..character,
    playing = true
  }
  
  local params = 
  {
    normal = normal,
    cage   = cell,
    charm  = cell
  }
  
  ObjSet( anm_name, params[ state ] );

end;
--------------------------------------------------------------------------------
-- fucntion .. () end;
--------------------------------------------------------------------------------
function private.Animate_Action ( character, state, func_after )

  local normal = 
  {
    action = "action",
    static = "static"
  };
  
  local cell =
  {
    action = "action_"..character,
    static = "static_"..character
  };
  
  local animfunc =
  {
    normal = normal,
    cage   = cell,
    charm  = cell
  };
  
  local anm_name = "anm_map_characters_icon_"..character;
  
  AnimPlay( anm_name, animfunc[ state ].action, function ()

    AnimPlay( anm_name, animfunc[ state ].static, "" );
    if ( func_after ) then func_after(); end;
  
  end );

end;
--------------------------------------------------------------------------------
-- fucntion .. () end;
--------------------------------------------------------------------------------
function private.GetAnimData_QuestStart ( character, quest )

  local obj_name = "obj_map_characters_"..character;
  local anm_name = "anm_map_characters_icon_"..character;

  local states = 
  {
    normal = -- Character in normal state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        private.Indicator_Enable( character, true );
        private.Animate_Action( character, "normal" );
        
        if ( func_after ) then func_after(); end;
      
      end
    },
    cage = -- Character in cage
    {
      name        = obj_name,
      res         = "assets/content/map/effects/anm_appearance",
      coords      = { pos_x = 0, pos_y = 0 },
      func_dummy  = function ()
      
        private.Sound_Play( "metamorphose" );
        private.Animate_Action( character, "cage" );
      
      end,
      func_action = function ( func_after )
      
        private.Indicator_Enable( character, true );
        if ( func_after ) then func_after(); end;
      
      end
    },
    charm = -- Transform character to charmed state
    {
      name        = obj_name,
      skip_focus  = true,
      func_action = function ( func_after )
      
        local enemy = m_characters.GetMainEnemy( character );
        local obj_enemy = "anm_map_point_enemy_"..enemy;
        
        -- Enemy scaling hack
        local scales =
        {
          [ "trollking_stage_330" ] = 0.6
        }
        
        v_map.Focus( obj_enemy,
        {
          scale      = scales[ enemy ] or 1.0,
          func_after = function ()
          
            local obj_parent = ObjGetRelations( obj_enemy ).parent;
            ObjClone( obj_enemy, "_dummy", 1, obj_parent );

            AnimPlay( obj_enemy.."_dummy", "dummy", function ()
            
              ObjDelete( obj_enemy.."_dummy" );
              v_map.Focus( anm_name,
              {
                func_after = function ()
                
                  private.SetVisualState( character, "charm" );
                  private.Sound_Play( "metamorphose" );
                  
                  AnimPlay( anm_name, "catch", function ()
                  
                    AnimPlay( anm_name, "static_"..character, "" );
                    private.Indicator_Enable( character, true );
                    
                    if ( func_after ) then func_after(); end;
                    
                  end );

                end
              } );
              
            end );

            AnimPlay( obj_enemy, "action", function ()
              AnimPlay( obj_enemy, "static", "" );
            end );
            local section = m_map.Section_Current_Name();
            cmn.PlaySnd( "assets/content/"..section.."/aud_boss_attack" );
          end
        } );
      
      end
    }
  }
  
  local state = private.GetVisualState_Active( character, quest.name );
  return states[ state ];

end;
--------------------------------------------------------------------------------
function private.GetAnimData_ItemReceive ( character, quest )

  local obj_name = "obj_map_characters_"..character;
  local anm_name = "anm_map_characters_icon_"..character;

  local states = 
  {
    normal = -- Character in normal state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        AnimPlay( anm_name, "action", function ()
          AnimPlay( anm_name, "static", "" );
        end );
        
        if ( func_after ) then func_after(); end;
      
      end
    },
    cage = -- Character in cage
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        AnimPlay( anm_name, "action_"..character, function ()
          AnimPlay( anm_name, "static_"..character, "" );
        end );
        
        if ( func_after ) then func_after(); end;
      
      end
    },
    charm = -- Character in charmed state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        AnimPlay( anm_name, "action_"..character, function ()
          AnimPlay( anm_name, "static_"..character, "" );
        end );
        
        if ( func_after ) then func_after(); end;
      
      end
    }
  }
  
  local state = private.GetVisualState_Active( character, quest.name );
  return states[ state ];

end;
--------------------------------------------------------------------------------
function private.GetAnimData_QuestDone ( character, quest )

  local obj_name = "obj_map_characters_"..character;
  
  local state = private.GetVisualState_Active( character, quest.name );
  
  local free_anim =
  {
    name        = obj_name,
    res         = "assets/content/map/effects/anm_appearance_character",
    coords      = { pos_x = 0, pos_y = 0 },
    func_preset = function ()
    
      private.Indicator_Disable( character, true );
      private.Animate_Action( character, state );
    
    end,
    func_dummy  = function ()
    
      private.SetVisualState( character, "normal" );
      private.Animate_Action( character, "normal" );
      private.Sound_Play( "metamorphose" );
    
    end,
    func_action = function ( func_after )
      
      if ( func_after ) then func_after(); end;
    
    end
  }

  local states = 
  {
    normal = -- Character in normal state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        private.Indicator_Disable( character, true );
        private.Animate_Action( character, "normal", func_after );
        
      end
    },
    cage  = free_anim, -- Transform cage to character
    charm = free_anim  -- Transform character from charmed to normal state
  }
  
  return states[ state ];

end;
--------------------------------------------------------------------------------
function public.Callback_AnimateAppear ( build_name )

  local character = m_characters.GetCharacterByBuild( build_name );
  if  ( character ) then
  
    local character_name = character.name;
  
    local obj_name = "obj_map_characters_"..character_name;
    local anm_name = "anm_map_characters_icon_"..character_name;
  
    local anim  = {};
    anim.name   = obj_name;
    anim.res    = "assets/content/map/effects/anm_appearance";
    anim.coords = { pos_x = 0, pos_y = 0 };
    
    anim.func_preset = function ()
    
      private.SetUnlocked( character_name );
      ObjSet( obj_name, { alp = 0.0 } );
    
    end;
    
    anim.func_dummy = function ()
    
      private.Sound_Play( "lighting_app" );
      ObjSet( obj_name, { alp = 1.0 } );
      
      local state = private.GetVisualState_Unlocked( character_name );
      private.Animate_Action( character_name, state );
    
    end;
    
    anim.func_action = function ( func_after )
      
      if ( func_after ) then func_after(); end;
    
    end;
    
    return v_map.Callback_Animate( anim );
    
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------